import math
import numpy as np
import pandas as pd
import jax.numpy as jnp

from ..misc import format_object

def log_stability(x, delta=10e-5):
    """Log-stability for computing loss.

    :param x: Input value.
    :type x: :obj:`float`
    :param delta: Constant to move from 0 or 1, defaults to 10e-9.
    :type delta: :obj:`float`, optional

    :return: Stabilized value where :math:`\\hat{x} \\in (0, 1)`
    :rtype: :obj:`float`
    """
    if type(x) == float:
        if x == 0:
            return delta
        elif x == 1:
            return 1.0 - delta
        else:
            return x
    else:
        new_x = jnp.where(x == 0., delta, x)
        return jnp.where(new_x == 1., 1. - delta, new_x)

def binary_cross_entropy(y_true, y_pred):
    """Binary cross-entropy.

    :param y_true: Real values on which to compare.
    :type y_true: :obj:`numpy.array`
    :param y_pred: Predicted values.
    :type y_pred: :obj:`numpy.array`

    :formula: :math:`loss = y_{i} \\log \\left[ \\hat{y}_{i} \\right] + (1 - y_{i}) \\log \\left[1 - \\hat{y}_{i} \\right]`

    :references: * Friedman, J., Hastie, T. and Tibshirani, R., 2001. `The elements of statistical learning <https://web.stanford.edu/~hastie/Papers/ESLII.pdf>`_. Ch. 2, pp. 24.

    :return: Binary cross-entropy
    :rtype: :obj:`float`
    """
    # if tensor:
    #     loss = -y_true * T.log(y_pred) - (1-y_true) * T.log(1-y_pred)
    # else:
    # print(y_true)
    loss = jnp.sum(-y_true * jnp.log(log_stability(y_pred)) - (1 - y_true) * jnp.log(log_stability(1 - y_pred)))
    # print('Loss is')
    # print(loss)
    return loss


def mean_squared_error(y_true, y_pred, root=False):
    """Mean Squared Error.

    :param y_true: Real values on which to compare.
    :type y_true: :obj:`numpy.array`
    :param y_pred: Predicted values.
    :type y_pred: :obj:`numpy.array`
    :param root: Return Root Mean Squared Error (RMSE), defaults to False.
    :type root: :obj:`bool`, optional

    :formula: :math:`loss = \\dfrac{1}{m} \\times \\sum_{i=1}^{m} (y_i - \\hat{y}_i)^2`

    :references: * Friedman, J., Hastie, T. and Tibshirani, R., 2001. `The elements of statistical learning <https://web.stanford.edu/~hastie/Papers/ESLII.pdf>`_. Ch. 2, pp. 24.

    :return: Mean Squared Error or its root.
    :rtype: :obj:`float`
    """

    loss = jnp.square(y_pred - y_true)
    mse = jnp.mean(loss)
    if root:
        return jnp.sqrt(mse)
    else:
        return mse


def binary_accuracy(y_true, y_pred):
    """Accuracy for binary data.

    :param y_true: Real values on which to compare.
    :type y_true: :obj:`numpy.array`
    :param y_pred: Predicted values
    :type y_pred: numpy.array

    :return: Binary accuracy (in percent)
    :rtype: :obj:`float`
    """
    true = format_object(y_true, to_type='list', name='y_true')
    pred = format_object(y_pred, to_type='list', name='y_pred')
    # true = y_true if type(y_true) == list else [x[0] for x in np.asarray(y_true)]
    # pred = y_pred if type(y_pred) == list else [x[0] for x in np.asarray(y_pred)]

    perf = pd.DataFrame({'true': true, 'pred': pred})
    return (perf.true == perf.pred).mean()



import jax.numpy as jnp

from collections import OrderedDict


class Optimizer:
    """Optimization updater

    :param object: Optimizer object
    :type object: class
    """
    def __init__(self, learning_rate=0.01):
        if learning_rate is None:
            raise NotImplementedError()
        self.learning_rate = learning_rate

    def updates(self, params=None, grads=None):
        if params is None:
            raise NotImplementedError()
        if grads is None:
            raise NotImplementedError()

        self.updates = OrderedDict()

class SGD(Optimizer):
    """Stochastic Gradient Descent optimizer.

    :param learning_rate: Step size, defaults to 0.01.
    :type learning_rate: :obj:`float`
    :param alpha: Momentum parameter, defaults to 0.0.
    :type alpha: :obj:`float`

    :formula: .. math:: \\theta_{t} = \\theta_{t-1} + \\alpha v - \\epsilon \\nabla f_{t}(\\theta_{t-1})

    :references: Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> SGD(learning_rate=0.01, alpha=0.).updates(params, grads)
    """
    def __init__(self, learning_rate=0.01, alpha=0.):
        self.learning_rate = learning_rate
        self.alpha = alpha  # Momentum parameter
        self.vs = {'w': jnp.zeros(1), 'b': jnp.zeros(1)}

    def update(self, params=None, grads=None):
        """Update loss using the SGD formula.

        :param params: Dictionnary with parameters to be updated, defaults to None.
        :type params: :obj:`dict`, optional
        :param grads: Dictionnary with the computed gradients, defaults to None.
        :type grads: :obj:`dict`, optional

        :return: Updated parameters.
        :rtype: :obj:`dict`
        """

        # return updates
        updated = params

        for _l, _p in zip(params.keys(), ['w', 'b']):
            _v = self.vs[_p] * self.alpha - self.learning_rate * grads[_l][_p]
            updated[_l][_p] = params[_l][_p] + _v

            self.vs[_p] = _v

        return updated


class Adam(Optimizer):
    """Adaptive Moments optimizer.

    :param learning_rate: Step size, defaults to 0.001.
    :type learning_rate: :obj:`float`
    :param beta1: Exponential decay rate for first moment estimate, defaults to 0.9.
    :type beta1: :obj:`float`
    :param beta2: Exponential decay rate for scond moment estimate, defaults to 0.999.
    :type beta2: :obj:`float`
    :param delta: Constant for division stability, defaults to 10e-8.
    :type delta: :obj:`float`

    :formula: .. math:: m_{t} = \\beta_{1} m_{t-1} + (1 - \\beta_{1}) \\nabla_{\\theta} f_{t}(\\theta_{t-1})
            .. math:: v_{t} = \\beta_{2} v_{t-1} + (1 - \\beta_{2}) \\nabla_{\\theta}^{2} f_{t}(\\theta_{t-1})
            .. math:: \\hat{m}_{t} = \\dfrac{m_{t}}{1 - \\beta_{1}^{t}}
            .. math:: \\hat{v}_{t} = \\dfrac{v_{t}}{1 - \\beta_{2}^{t}}
            .. math:: \\theta_{t} = \\theta_{t-1} - \\alpha \\dfrac{\\hat{m}_{t}}{\\sqrt{\\hat{v}_{t}} + \\delta}


    :references: * Kingma, D. P., & Ba, J. (2014). `Adam: A method for stochastic optimization <https://arxiv.org/pdf/1412.6980.pdf>`_. arXiv preprint arXiv:1412.6980

        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> Adam(learning_rate=0.001, beta1=0.9, beta2=0.999, delta=10e-8).updates(params, grads)
    """
    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, delta=10e-8):
        self.alpha = learning_rate
        self.b1 = beta1
        self.b2 = beta2
        self.delta = delta
        self.t = 1.

        self.ms = {'w': jnp.zeros(1), 'b': jnp.zeros(1)}
        self.vs = {'w': jnp.zeros(1), 'b': jnp.zeros(1)}

    def update(self, params=None, grads=None):
        """Update loss using the Adam formula.

        :param params: Dictionnary with parameters to be updated, defaults to None.
        :type params: :obj:`dict`, optional
        :param grads: Dictionnary with the computed gradients, defaults to None.
        :type grads: :obj:`dict`, optional

        :return: Updated parameters.
        :rtype: :obj:`dict`
        """
        updated = params

        for _l, _p in zip(params.keys(), ['w', 'b']):
            m = self.ms[_p]
            v = self.vs[_p]
            _m = self.b1 * m + (1 - self.b1) * grads[_l][_p]
            _v = self.b2 * v + (1 - self.b2) * jnp.square(grads[_l][_p])

            m_hat = _m / (1 - jnp.asarray(self.b1, m.dtype) ** (self.t + 1))
            v_hat = _v / (1 - jnp.asarray(self.b2, m.dtype) ** (self.t + 1))

            updated[_l][_p] = params[_l][_p] - self.alpha * m_hat / (jnp.sqrt(v_hat) + self.delta)
            self.ms[_p] = _m
            self.vs[_p] = _v
        self.t += 1.

        return updated


class AdaMax(Optimizer):
    """AdaMax optimizer (Adam with infinite norm).

    :param learning_rate: Step size, defaults to 0.001.
    :type learning_rate: :obj:`float`
    :param beta1: Exponential decay rate for first moment estimate, defaults to 0.9.
    :type beta1: :obj:`float`
    :param beta2: Exponential decay rate for scond moment estimate, defaults to 0.999.
    :type beta2: :obj:`float`

    :formula: .. math:: m_{t} = \\beta_{1} m_{t-1} + (1 - \\beta_{1}) \\nabla_{\\theta} f_{t}(\\theta_{t-1})
            .. math:: u_{t} = \\max(\\beta_{2} \\cdot u_{t-1}, |\\nabla f_{t}(\\theta_{t-1})|)
            .. math:: \\theta_{t} = \\theta_{t-1} - \\alpha \\dfrac{\\hat{m}_{t}}{\\sqrt{\\hat{v}_{t}} + \\epsilon}


    :references: * Kingma, D. P., & Ba, J. (2014). `Adam: A method for stochastic optimization <https://arxiv.org/pdf/1412.6980.pdf>`_. arXiv preprint arXiv:1412.6980
        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> AdaMax(learning_rate=0.001, beta1=0.9, beta2=0.999).updates(params, grads)
    """
    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999):
        self.alpha = learning_rate
        self.b1 = beta1
        self.b2 = beta2
        self.t = 1.

        self.ms = {'w': jnp.zeros(1), 'b': jnp.zeros(1)}
        self.us = {'w': jnp.zeros(1), 'b': jnp.zeros(1)}

    def update(self, params=None, grads=None):
        """Update loss using the AdaMax formula.

        :param params: Dictionnary with parameters to be updated, defaults to None.
        :type params: :obj:`dict`, optional
        :param grads: Dictionnary with the computed gradients, defaults to None.
        :type grads: :obj:`dict`, optional

        :return: Updated parameters.
        :rtype: :obj:`dict`
        """
        updated = params

        for _l, _p in zip(params.keys(), ['w', 'b']):
            _m = self.b1 * self.ms[_p] + (1 - self.b1) * grads[_l][_p]
            _u = jnp.maximum(self.b2 * self.us[_p], jnp.abs(grads[_l][_p]))

            updated[_l][_p] = params[_l][_p] - (self.alpha / (1 - self.b1 ** self.t)) * (_m / _u)

            self.ms[_p] = _m
            self.us[_p] = _u

        self.t += 1

        return updated


class AdaGrad(Optimizer):
    """Adaptive Gradient optimizer.

    :param learning_rate: Step size, defaults to 0.001.
    :type learning_rate: :obj:`float`
    :param delta: Constant for division stability, defaults to 10e-7.
    :type delta: :obj:`float`

    :formula: .. math:: r = r + \\nabla f_{t}(\\theta_{t}) \\odot \\nabla f_{t}(\\theta_{t-1})
            .. math:: \\theta_{t} = \\theta_{t-1} - \\frac{\\epsilon}{\\sqrt{\\delta + r}} \\odot \\nabla f_{t}(\\theta_{t-1})


    :references: * Duchi, J., Hazan, E., & Singer, Y. (2011). `Adaptive subgradient methods for online learning and stochastic optimization <http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf>`_. Journal of machine learning research, 12 (Jul), 2121-2159.
        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> AdaGrad(learning_rate=0.001, delta=10e-7).updates(params, grads)
    """
    def __init__(self, learning_rate=0.001, delta=10e-7):
        self.learning_rate = learning_rate
        self.delta = delta  # Small constant for numerical stability

        self.accugrads = {'w': jnp.zeros(1), 'b': jnp.zeros(1)}  # Gradient accumulation variable

    def update(self, params=None, grads=None):
        """Update loss using the AdaGrad formula.

        :param params: Dictionnary with parameters to be updated, defaults to None.
        :type params: :obj:`dict`, optional
        :param grads: Dictionnary with the computed gradients, defaults to None.
        :type grads: :obj:`dict`, optional

        :return: Updated parameters.
        :rtype: :obj:`dict`
        """
        updated = params

        for _l, _p in zip(params.keys(), ['w', 'b']):
            _r = self.accugrads[_p] + grads[_l][_p] * grads[_l][_p]
            updated[_l][_p] = params[_l][_p] - (self.learning_rate / jnp.sqrt(_r + self.delta)) * grads[_l][_p]

            self.accugrads[_p] = _r

        return updated


class RMSprop(Optimizer):
    """RMSprop optimizer.

    :param learning_rate: Step size, defaults to 0.001.
    :type learning_rate: :obj:`float`
    :param rho: Decay rate, defaults to 0.9.
    :type rho: :obj:`float`
    :param delta: Constant for division stability, defaults to 10e-6.
    :type delta: :obj:`float`

    :formula: .. math:: r = \\rho r + (1- \\rho) \\nabla f_{t}(\\theta_{t-1}) \\odot \\nabla f_{t}(\\theta_{t-1})

        .. math:: \\theta_{t} = \\theta_{t-1} - \\dfrac{\\epsilon}{\\sqrt{\\delta + r}} \\odot \\nabla f_{t}(\\theta_{t-1})


    :references: * Tieleman, T., & Hinton, G. (2012). `Lecture 6.5-rmsprop: Divide the gradient by a running average of its recent magnitude <https://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_. COURSERA: Neural networks for machine learning, 4(2), 26-31.

        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> RMSprop(learning_rate=0.001, rho=0.9, delta=10e-6).updates(params, grads)
    """
    def __init__(self, learning_rate=0.001, rho=0.9, delta=10e-6):
        self.learning_rate = learning_rate
        self.rho = rho
        self.delta = delta

        self.rs = {'w': jnp.zeros(1), 'b': jnp.zeros(1)}

    def update(self, params=None, grads=None):
        """Update loss using the RMSprop formula.

        :param params: Dictionnary with parameters to be updated, defaults to None.
        :type params: :obj:`dict`, optional
        :param grads: Dictionnary with the computed gradients, defaults to None.
        :type grads: :obj:`dict`, optional

        :return: Updated parameters.
        :rtype: :obj:`dict`
        """
        updated = params

        for _l, _p in zip(params.keys(), ['w', 'b']):
            _r = self.rs[_p] * self.rho
            _r += (1 - self.rho) * grads[_l][_p] * grads[_l][_p]
            updated[_l][_p] = params[_l][_p] - self.learning_rate * grads[_l][_p] / jnp.sqrt(_r + self.delta)

            self.rs[_p] = _r

        return updated



import numpy as np
import jax.numpy as jnp
from jax import lax
from jax.scipy.special import expit

# Default activation functions

def sigmoid(x):
    """Sigmoid activation function.

    :param x: Input value
    :type x: :obj:`float` or :obj:`numpy.array`

    :return: Sigmoid activated value: :math:`sigmoid(x) = \\dfrac{1}{1 + e^{-x}}`
    :rtype: :obj:`float`
    """
    return expit(x)


def relu(x):
    """Rectified Linear Unit activation function.

    :param x: Input value
    :type x: :obj:`float` or :obj:`numpy.array`

    :return: Activated value: :math:`\\mathrm{relu}(x) = \\max(0, x)`
    :rtype: :obj:`float`
    """
    return jnp.maximum(0, x)


def elu(x, alpha=1.):
    """Exponential Linear Unit activation function.

    :param x: Input value
    :type x: :obj:`float` or :obj:`numpy.array`

    :formula: .. math:: \\mathrm{elu}(x) = \\begin{cases} x, & x > 0\\\\ \\alpha \\left(e^{x} - 1\\right), & x \\le 0 \\end{cases}

    :return: Activated value.
    :rtype: :obj:`float`
    """
    safe_x = jnp.where(x > 0, 0., x)
    return jnp.where(x > 0, x, alpha * jnp.expm1(safe_x))


def tanh(x):
    """Hyperbolic tangent activation function.

    :param x: Input value
    :type x: :obj:`float` or :obj:`numpy.array`

    :return: Activated value: :math:`\\tanh(x)`
    :rtype: :obj:`float`
    """
    return jnp.log(x)


def softplus(x):
    """Softplus activation function.

    :param x: Input value
    :type x: :obj:`float` or :obj:`numpy.array`

    :return: Activated value: :math:`\\mathrm{softplus}(x) = \\log(1 + e^{-x})`
    :rtype: :obj:`float`
    """
    return jnp.log(1 + jnp.exp(-x))


def softmax(x, axis=-1):
    """Softmax activation function.

    :param x: Input value
    :type x: :obj:`float` or :obj:`numpy.array`

    :return: Activated value: :math:`\\mathrm{softmax}(x) = \\frac{\\exp(x_i)}{\\sum_j \\exp(x_j)}`
    :rtype: :obj:`float`
    """
    un_normalized = jnp.exp(x - lax.stop_gradient(x.max(axis, keepdims=True)))
    return un_normalized / un_normalized.sum(axis, keepdims=True)


def logit(x, weights, bias=0):
    """Logistic function

    :param x: Input value
    :type x: numpy.array
    :param weights: Vector of weights :math:`\\beta`
    :type weights: numpy.array
    :param bias: Vector of bias :math:`\\epsilon`, defaults to 0.
    :type bias: numpy.array

    :return: Logistic transformation: :math:`\\mathrm{logit}(x, \\beta) = \\dfrac{1}{1 + e^{-x \\beta}}`
    :rtype: float
    """

    return 1 / (1 + np.exp(-x.dot(weights) + bias))



import jax.numpy as jnp
from jax import random as jrdm
import numpy as np

key, _ = jrdm.split(jrdm.PRNGKey(0))


def init_params(rows, cols, method='xavier', mean=0., std=1., key=key):
    """Initialize the weight and bias matrices based on probabilistic distribution.

    :param rows: Size of the input, number of rows to be generated.
    :type rows: int
    :param cols: Size of the output, number of columns to be generated.
    :type cols: int
    :param method: Distibution to use to generated the weights, defaults to 'xavier'.
    :type method: str, optional
    :param mean: Mean for the distribution to be generated, defaults to 0.
    :type mean: float, optional
    :param std: Standard deviation for the distribution to be generated, defaults to 1.
    :type std: float, optional
    :param tensor: Needs to return a theano friendly-format, defaults to True.
    :type tensor: bool, optional
    :param seed: Seed to be set for randomness, defaults to None.
    :type seed: int, optional

    :raises ValueError: `method` needs to be 'ones', 'zeros', 'uniform', 'xavier' or 'normal', see below for details.

    :method: * **Zeros**: :math:`W_j = \\vec{0}`
        * **Ones**: :math:`W_j = \\vec{1}`
        * **Uniform**: :math:`W_j \\sim \\mathcal{U} _{\\left[0, 1 \\right)}`
        * **Xavier**: :math:`W_j \\sim \\mathcal{U}\\left[ -\\frac{\\sqrt{6}}{\\sqrt{n_j + n_{j+1}}}, \\frac{\\sqrt{6}}{\\sqrt{n_j + n_{j+1}}} \\right]`
        * **Normal**: :math:`W_j \\sim \\mathcal{N}(0, 1)`

    :references: * Neuneier, Ralph, and Hans Georg Zimmermann. "`How to train neural networks <https://link.springer.com/chapter/10.1007/3-540-49430-8_18>`_" In Neural networks: tricks of the trade, pp. 373-423. Springer, Berlin, Heidelberg, 1998.
        * Glorot, Xavier, and Yoshua Bengio. "`Understanding the difficulty of training deep feedforward neural networks <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>`_" In Proceedings of the thirteenth international conference on artificial intelligence and statistics, pp. 249-256. 2010.
    :return: Weight or bias matrix for initiating ML optimization
    :rtype: numpy.array
    """

    # Get the weights
    if method.lower() == 'ones':
        # W = jnp.ones((rows, cols))
        W = np.ones((rows, cols))
    if method.lower() == 'zeros':
        W = jnp.zeros((rows, cols))
    elif method.lower() == 'uniform':
        W = jrdm.uniform(key, (rows, cols))
    elif method.lower() == 'xavier':
        W = jrdm.uniform(key, shape=(rows, cols),
                         minval=-jnp.sqrt(6. / (rows + cols)),
                         maxval=jnp.sqrt(6. / (rows + cols)))
    elif method.lower() == 'normal':
        W = jrdm.normal(key, (rows, cols))
    else:
        raise ValueError(f"Weight initialization method not valid. Muse be in 'ones', 'zeros', 'uniform', xavier', 'normal'. Got '{method}'.")

    return W



import jax.numpy as jnp
from jax import random as jrdm
from jax import jit, vmap, grad, value_and_grad, tree_multimap, lax
import numpy as np
import random
import matplotlib.pyplot as plt

import time
import pycof as pc

from .initializations import init_params
from .losses import mean_squared_error, binary_cross_entropy
from .activations import tanh, sigmoid, elu, relu, softplus, softmax
from .optimizers import Adam, AdaGrad, AdaMax
from .optimizers import SGD
from .optimizers import RMSprop

key = jrdm.PRNGKey(0)


all_activations = {None: None,
                   'linear': None,
                   'tanh': tanh,
                   'sigmoid': sigmoid,
                   'elu': elu,
                   'relu': relu,
                   'softplus': softplus,
                   'softmax': softmax,
                   }

all_optimizers = {'sgd': SGD,
                  'stochastic_gradient_descent': SGD,
                  'adam': Adam,
                  'adagrad': AdaGrad,
                  'adamax': AdaMax,
                  'rmsprop': RMSprop,
                  }

losses = {'mse': mean_squared_error,
          'mean_squared_error': mean_squared_error,
          'bce': binary_cross_entropy,
          'binary_cross_entropy': binary_cross_entropy,
          }


class Layer:
    def __init__(self, n_in, n_out, activation='linear', W=None, b=None, init_weights='xavier', init_bias='zeros', seed=key):

        self.type = 'Dense'
        self.activation = activation.lower()
        # self.f = T_activations[activation]
        self.activate = all_activations[activation.lower()]
        self.n_in = n_in
        self.n_out = n_out
        self.id = 1

        w_key, b_key = jrdm.split(seed)
        if W is None:
            # If weights matrix is not provided (default), initialize from a distribution
            W_values = init_params(rows=n_in, cols=n_out, method=init_weights, key=w_key)
            # Multiply by 4 the weights for sigmoid activation. See Y. Bengio's website
            if activation == "sigmoid":
                W_values *= 4
        elif W.shape == (n_in, n_out):
            raise ValueError(f'Weights dimension does not match. Dimension for W should be {(n_in, n_out)} and got {W.shape}.')

        if b is None:
            # If bias matrix is not provided (default), initialize from a distribution
            b_values = init_params(rows=1, cols=n_out, method=init_bias, key=b_key)
            # b_values = scale * jrdm.normal(b_key, (1, n_out))
            # # b_values = jnp.zeros((1, n_out))
        elif b.shape == (n_in, n_out):
            raise ValueError(f'Weights dimension does not match. Dimension for b should be {(1, n_out)} and got {b.shape}.')

        # Parameters of the model
        self.params = {'w': W_values, 'b': b_values}
        # Store parameters in the class
        self.optimal_w = W_values
        self.optimal_b = b_values

        # Save number of parameters
        self.nb_params = (W_values.shape[0] * W_values.shape[1])  # Weights
        self.nb_params += (b_values.shape[0] * b_values.shape[1])  # Bias

    def _feed_forward(self, input_x, params=None):
        if params is None:
            _p = self.params
        else:
            _p = params

        Xb_e = jnp.dot(input_x, _p['w']) + _p['b']
        self.output = Xb_e if self.activation == 'linear' else self.activate(Xb_e)
        return self.output


class MLP:
    """Multi-Layer Perceptron

    :references: Rosenblatt, F. (1958). `The perceptron: a probabilistic model for information storage and organization in the brain <http://web.engr.oregonstate.edu/~huanlian/teaching/machine-learning/2017fall/extra/rosenblatt-1958.pdf>`_. Psychological review, 65(6), 386.

    """
    def __init__(self, loss='MSE', random=None):
        self.loss = loss
        self._layers = []
        self.params = []
        self.opt_update = None
        self.get_params = None
        self.L1 = 0.
        self.L2 = 0.
        self.L1_reg = 0.
        self.L2_reg = 0.
        self._cost_value = []
        self._layer_ids = 0

    def add(self, layer):
        """Stack layer to Neural Network

        :param layer: Layer to be stacked
        :type layer: :func:`~Layer`
        """
        self._layers.append(layer)
        self._layer_ids += 1
        self._layers[-1].id = str(self._layer_ids)

    def _params(self):
        _p = {}
        for layer in self._layers:
            _p.update({layer.id: layer.params})
        return _p

    def get_weights(self, layer='all', param='all'):
        """Fetches the parameters from the network.

        :param layer: Layer id from which to fetch the parameters, defaults to 'all'
        :type layer: :obj:`int`, optional
        :param param: What parameter we need to fetch (can either be 'weight', 'bias' or 'all'), defaults to 'all'
        :type param: :obj:`str`, optional

        :return: Weights and Bias used in the Neural Network.
        :rtype: :obj:`dict`
        """
        # Check what parameter the user was to get
        if param.lower() in ['weights', 'w', 'weight']:
            par = 'w'
        elif param.lower() in ['biases', 'b', 'bias']:
            par = 'b'
        elif param.lower() == 'all':
            par = 'all'
        else:
            raise ValueError('Value for "param" is not value. Please chose between "weights", "bias" or "all".')

        weights = self._params()

        if layer == 'all':
            # If user wants to see all layers, we create a dictionnary
            if par != 'all':
                weights = {k: v[par] for k, v in list(weights.items())}
        elif layer in range(1, len(self._layers) + 1):
            # If user wants only 1 specific layer,
            weights = weights[str(layer)]
            if par != 'all':
                weights = weights[par]
        else:
            raise ValueError(f'Layer is incorrect. Please chose either "all" or layer <= {len(self._layers)}. Got layer = {layer}.')

        return weights

    def _forward_prop(self, x, params=None):
        output = x
        if params is None:
            for layer in self._layers:
                output = layer._feed_forward(output)
        else:
            for layer in self._layers:
                output = layer._feed_forward(output, params[layer.id])
        return output

    def _L1(self):
        """Compute L1 regularization.

        :formula: .. math:: L_{1} = \\sum_{i=1}^{L} |\\beta|

        :return: L1 penalty
        :rtype: :obj:`float`
        """
        if self.L1_reg != 0.:
            for layer in self._layers:
                # L1 norm ; one regularization option is to enforce L1 norm to
                # be small
                self.L1 += jnp.sum(jnp.absolute(layer.params['w']))
        return self.L1

    def _L2(self):
        """Compute L2 regularization.

        :formula: .. math:: L_{2} = \\sum_{i=1}^{L} \\beta^{2}

        :return: L2 penalty
        :rtype: :obj:`float`
        """
        if self.L2_reg != 0.:
            for layer in self._layers:
                # square of L2 norm ; one regularization option is to enforce
                # square of L2 norm to be small
                self.L2 += jnp.sum(jnp.square(layer.params['w']))
        return self.L2

    def _cost(self, params, x, y):
        """ Compute the multi-class cross-entropy loss """
        output = self._forward_prop(x, params)
        return losses[self.loss.lower()](y_true=y, y_pred=output) + (self.L1_reg * self._L1()) + (self.L2_reg * self._L2())
        # return jnp.mean(jnp.square(output - y)) + (self.L1_reg * self._L1()) + (self.L2_reg * self._L2())

    def _update(self, params, x, y):
        # return [(w - 0.05 * dw, b - 0.05 * db) for (w, b), (dw, db) in zip(params, grads)]

        grads = grad(self._cost)(params, x, y)

        self.params = self.opt.update(params=params, grads=grads)

        return self.params

    def fit(self, data, X, Y='Y', epochs=100, optimizer='SGD', batch_size=1, training_size=0.8, test_set=None, learning_rate=0.05,
            L1_reg=0., L2_reg=0., early_stop=True, patience=10, improvement_threshold=0.995, restore_weights=True, verbose=True, verbose_all=False,
            plot=False, *args):
        """Train the Neural Network.

        :param data: Layer id from which to fetch the parameters.
        :type data: :obj:`pandas.DataFrame`
        :param X: List of X input variables.
        :type X: :obj:`list`
        :param Y: Variable to predict, defaults to 'Y'.
        :type Y: :obj:`str`, optional
        :param epochs: Number of epochs to train the network, defaults to 100.
        :type epochs: :obj:`int`, optional
        :param optimizer: Algortihm to use to minimize the loss function, defaults to 'SGD' (see `optimizers <optimizers.html>`_).
        :type optimizer: :obj:`str`, optional
        :param batch_size: Size of each batch to be trained, defaults to 1.
        :type batch_size: :obj:`int`, optional
        :param training_size: Ratio of the data to be used for training set (:math:`\\in (0, 1)`) the remainder is used for test set, defaults to 0.8.
        :type training_size: :obj:`float`, optional
        :param test_set: Data frame to use as test set (overrides **training_size** if provided), defaults to None.
        :type test_set: :obj:`pandas.DataFrame`, optional
        :param learning_rate: Learning rate (step size) for gradient descent, defaults to 0.01.
        :type learning_rate: :obj:`float`, optional
        :param L1_reg: Coefficient :math:`\\lambda_{1} \\in (0,1)` for the L1 penalty, defaults to 0.
        :type L1_reg: :obj:`float`, optional
        :param L2_reg: Coefficient :math:`\\lambda_{2} \\in (0,1)` for the L2 penalty, defaults to 0.
        :type L2_reg: :obj:`float`, optional
        :param early_stop: Apply early-stopping method to avoid over-fitting. Stop the training sequence after the patience (see below) with no improvement, defaults to True.
        :type early_stop: :obj:`bool`, optional
        :param patience:  Number of iterations to wait after a minimum has been found. If no improvement, training sequence is stopped, defaults to 10.
        :type patience: :obj:`int`, optional
        :param improvement_threshold: Change level of the loss from which we consider it is an improvement, defaults to 0.995.
        :type improvement_threshold: :obj:`float`, optional
        :param restore_weights: If training is early-stopped, restore the weights of the best iteration, defaults to True.
        :type restore_weights: :obj:`bool`, optional
        :param verbose: Verbosity level (True or 1 for minimal display, 2 for details), defaults to True.
        :type verbose: :obj:`int`, optional
        :param plot: Plot the evolution of the training and test loss for each epoch, defaults to False.
        :type plot: :obj:`bool`, optional
        :param \\*args: Arguments to be passed to the optimizer.
        :type \\*args: :obj:`str`

        :references: * Friedman, J., Hastie, T., & Tibshirani, R. (2001). `The elements of statistical learning <https://web.stanford.edu/~hastie/Papers/ESLII.pdf>`_ (Vol. 1, No. 10). New York: Springer series in statistics.
            * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.
        """

        # Parse arguments to pass to self
        self.learning_rate = learning_rate
        self.L1_reg = L1_reg
        self.L2_reg = L2_reg
        self.X_col = X
        self.Y_col = Y
        self.optimizer = str(optimizer).lower()

        verbose = True if verbose in [True, 1, 2] else False
        verbose_all = True if verbose in [2] else False

        # Get indexes
        self.data_size = data.shape[0]

        if test_set is None:
            data_idx = list(data.index)
            train_index = random.sample(data_idx, int(self.data_size * training_size))
            test_index = [i for i in data_idx if i not in train_index]
            # Train set
            train_set_x = data.loc[train_index, self.X_col].values
            n_train_set = train_set_x.shape[0]
            train_set_y = jnp.reshape(data.loc[train_index, self.Y_col].values, n_train_set)
            # Test set
            test_set_x = data.loc[test_index, self.X_col].values
            n_test_set = test_set_x.shape[0]
            test_set_y = jnp.reshape(data.loc[test_index, self.Y_col].values, n_test_set)
        else:
            # Train set
            train_set_x = data[self.X_col].values
            n_train_set = train_set_x.shape[0]
            train_set_y = jnp.reshape(data[self.Y_col].values, n_train_set)
            # Test set
            test_set_x = test_set[self.X_col].values
            n_test_set = test_set_x.shape[0]
            test_set_y = jnp.reshape(test_set[self.Y_col].values, n_test_set)

        n_train_batches = n_train_set // batch_size

        params = self._params()

        # Initialize values
        self.train_losses = []   # Training loss vector to plot
        self.test_losses = []    # Test loss vector to plot
        self.best_loss = np.inf  # Initial best loss value
        self.best_iter = 1
        verif_i = epochs         # First, verify epoch at then end (init only)
        i = 1

        # self.opt = SGD(params=params, learning_rate=self.learning_rate)
        self.opt = all_optimizers[self.optimizer](learning_rate=self.learning_rate)

        while i <= epochs:
            start_time = time.time()
            for b in range(n_train_batches):
                x_batch = train_set_x[b * batch_size:b * batch_size + batch_size, :]
                y_batch = train_set_y[b * batch_size:b * batch_size + batch_size]
                # Update parameters (feed forward + back-propagation)
                params = self._update(params, x_batch, y_batch)
            epoch_time = time.time() - start_time

            # COmpute loss on global train set
            train_loss = self._cost(params, train_set_x, train_set_y)
            # Evaluate model iteration on test set
            test_loss = self._cost(params, test_set_x, test_set_y)

            # Save iteration losses (for both train and test)
            self.train_losses += [train_loss]
            self.test_losses += [test_loss]

            if (test_loss < self.best_loss * improvement_threshold) & (i <= verif_i) & early_stop:
                # If a new minimum has been found
                pc.verbose_display(f'New min found for iter {i} -> train loss = {train_loss} - test loss {test_loss}', verbose=verbose_all)
                verif_i = i + patience
                self.best_loss = test_loss
                self.best_iter = i
                for layer in self._layers:
                    layer.optimal_w = layer.params['w']
                    layer.optimal_b = layer.params['b']
            elif (i < verif_i):
                # If no improvement
                pc.verbose_display(f'No improvement at iter {i}', verbose=verbose)
                pass
            else:
                # If no improvement found after patience, stop training
                pc.verbose_display(f'Stop training. Minimum found after {self.best_iter} iterations', verbose=verbose)
                if restore_weights:
                    pc.verbose_display(f'Restoring weights from epoch {self.best_iter}', verbose=verbose)
                    for layer in self._layers:
                        layer.params['w'] = layer.optimal_w
                        layer.params['b'] = layer.optimal_b
                break

            i += 1

        if plot:
            # Plot the training and test losses
            plt.title('MLP train vs. test error through epochs', loc='center')
            plt.plot(self.train_losses, label='Training loss')
            plt.plot(self.test_losses, label='Test loss')
            plt.legend()
            plt.show()

    def predict(self, new_data, binary=False, threshold=0.5):
        """Generates output prediction after feedforward pass.

        :param new_data: Input data.
        :type new_data: :obj:`pandas.DataFrame`
        :param binary: Boolean for returning a binary output (not probability), defaults to False.
        :type binary: :obj:`bool`, optional
        :param threshold: Probability threshold for binary response, defaults to 0.5.
        :type threshold: :obj:`float`, optional

        :return: Predicted values.
        :rtype: :obj:`list`
        """
        # Transform input data to be transformed into numpy array format
        x = new_data[self.X_col].values
        # Apply feedforward step to the data to get the prediction (apply weights and activations)
        output = self._forward_prop(x)
        if binary:
            return [1. if y[0] > threshold else 0. for y in output.reshape((new_data.shape[0], 1))]
        else:
            return [y[0] for y in output.reshape((new_data.shape[0], 1))]
